import os
import random
from typing import Any, Dict, Optional, Tuple

import jwt
from jwt.api_jwt import decode_complete

#
# [HOW TO GENERATE KEYs]
# openssl genrsa -out private.pem 2048 # private key
# openssl rsa -in private.pem -out public.pem -RSAPublicKey_out # public key
#
# [LOCAL PEM DIR STRUCTURE]
# .pem
# └── stg
#     └── test
#         ├── private.pem
#         └── public.pem
#


def pem_random_key(env: str) -> Tuple[str, str]:
    keys_dir = f".pem/{env}"
    kid = random.choice(os.listdir(keys_dir))
    private_key = None
    with open(f"{keys_dir}/{kid}/private.pem") as f:
        private_key = f.read()
    return (kid, private_key)


def pem_encode(
    payload: Dict[str, Any],
    key: str,
    headers: Optional[Dict[str, Any]] = None,
) -> str:
    return jwt.encode(
        payload,
        key,
        algorithm="RS256",
        headers=headers,
    )


def pem_decode(token: str) -> Dict[str, Any]:
    completed = decode_complete(token, options={"verify_signature": False})
    public_key = open(
        f".pem/{completed['payload']['env']}/{completed['header']['kid']}/public.pem"
    ).read()
    return jwt.decode(token, public_key, algorithms=["RS256"])



from typing import Literal, Optional

from pydantic import BaseModel


class CheckoutModel(BaseModel):
    MerchantTradeNo: str
    TotalAmount: int
    ItemName: str = "Online Payment To Ecpay"
    TradeDesc: str = "no description"
    ChoosePayment: Literal["ALL", "Credit", "WebATM", "ATM", "CVS", "BARCODE"] = "ALL"
    BindingCard: Optional[Literal[0, 1]]
    MerchantMemberID: Optional[str]
    RelateNumber: Optional[str]
    PaymentType: str = "aio"
    InvoiceMark: Optional[Literal["Y", "N"]] = "N"
    # only fill this if InvoiceMark is Y
    TaxType: Optional[Literal["1", "2", "3", "9"]]
    ReturnURL: Optional[str]
    OrderResultURL: Optional[str]
    PaymentInfoURL: Optional[str]
    ClientRedirectURL: Optional[str]



from typing import Any, Dict, Tuple

from jwt.algorithms import get_default_algorithms
from jwt.api_jwt import decode, decode_complete

import plugin

from .pem_coder import pem_decode

# https://keys.revtel-api.com/certs.json
prod_cert = {
    "keys": [
        {
            "e": "AQAB",
            "kid": "bc36e546-f24a-4635-a77b-eecb20602504",
            "kty": "RSA",
            "n": "8oPMlKMnG9Uvyr8SMSLf5N7pl2ciEL-OWkAYngAzV1Y492cH8PWWFCibpiXU6iNlBbP_py2O6p8xfMHX1vGCF9uFU__iqc5RBbaCYL1kxczQbCt69tPLdyv3_6FNfMioc62Cym77rEVIa4uRLJl0TB_BJ89beCoL7BO6U1szGz3oVn-4igsc8GQRJOaZkZXY0JIBqB7dTHBgiq1R444ex_Tl1M6E9w45PxzowLOn1GWv6X5wyOSX1z-g60ErmDdxuu3N3lm2fJr9W-mjFgMvo4V9tBQmABaWO573I82IJymxt3C6B1g9Co9P1Thk5zMxtm33Om2FrizinUZbnQAtqQ",
        },
        {
            "e": "AQAB",
            "kid": "d4601977-6244-4a01-87dd-af0c1855886a",
            "kty": "RSA",
            "n": "yrIO_j-qbZfDlbLaTF-TQC38LVv8FVaXV4RPwNwSL_3QQA04HO1lnA8eXf2jV3RbI3WVey9jgDxjIidGAQQ0EJawXX7lPKWj56gDWBjPID7z-32DXicPww7vAgP8KD_GD00bwDnrnPCEveZQ9fAX-9zJpDNLRZlZ-GdgwOP55dIUFaOGeS1GP3jCBcy9auRzthBV4CO2Y_WzC69yL-4WgBPY_cIEc49vZ_Xyyo2mP36-fHzJc9eQ-moRczJ_hlCUe6DPfC99ojMFxXyd9SkD6BvTdnI5Bw-DPd4MCQj6KeSTuWTVfi2zsL2ooX0e8oLqmdhKqEB7T9LAr2efJkBGQQ",
        },
    ]
}

# https://keys.revtel-api.com/certs-stg.json
stg_cert = {
    "keys": [
        {
            "e": "AQAB",
            "kid": "df9885fa-942b-40c0-988c-9c3af76694bc",
            "kty": "RSA",
            "n": "x_-UkGLNQVRBXXW894iokDf7irGQ8rNPzsp-9N4ylVrkES93OL95E_rOdoK0Z8kCuwPOvKfc1QmgBFoJjMMaxI3zLDdDTl9QRIS1e1akecuAdzMj53X_t98Z2pgcT1paoDSkHh7qgYRKmt1xpU7fbKrogjdzqTv3vsnB3tQU2P1-9UJtH3-1BoAVhyiusFqXLH0o6Rp7drMbVYbvyj19nRwcBZz9gtO_bWyYGz0KUtFkm_vc31JmCARif7Tb4vc6FsmjGCgaQ9OSbJLgmYS7ZeVLFomyLKZuDeyAbS0rfzjC6Cf5heu6F2F44MdRoq-QK88nZ19fOpcJk1CkYPtySQ",
        },
        {
            "e": "AQAB",
            "kid": "1a7f2fbe-5112-4450-b7f1-27187d6030fb",
            "kty": "RSA",
            "n": "nPPHuzGJ8M9eZVr2f_CUrzFzyPQ0Ks9R31abO2B6qSOKQb_7aLQC7kOB02wWckyqpKhMRHTVbpKBJXYI1sga_iAaFfDyJ8-RVH3-hbpF1-_Bv7AteGJSZe2Etyi4kFXSZs2pDNOgUS6zvrkdQTlIqSst6MGJNKiaF1OpsmYlwAzJF37YAbrkuNOC1nbQorKkqQzSDa9667ZiEGoU65TGWP0FWwuzSBJGb8AOVNIEdIUaEdTrimoENzJOJR-RHeRGPLpU_Fe7TRj_RLyixDB6Hp_ZyYHW3Su7N3YBu2vZnp42d29E1UMFCU6k_5uoWR6rvosxCL1FkfaoXvpQ78oDmQ",
        },
    ]
}


def decode_static(token: str, pem: bool = False) -> Dict[str, Any]:
    if pem:
        return pem_decode(token)
    unverified = decode_complete(token, options={"verify_signature": False})
    signing_key = None
    if unverified["payload"]["env"] == "stg":
        plugin_resp: Tuple[bool, Any] = (
            getattr(plugin, "stg_cert", None) and plugin.stg_cert() or (False, None)
        )
        if plugin_resp[0]:
            signing_keys = plugin_resp[1]
        else:
            signing_keys = stg_cert
    else:
        plugin_resp = (
            getattr(plugin, "prod_cert", None) and plugin.prod_cert() or (False, None)
        )
        if plugin_resp[0]:
            signing_keys = plugin_resp[1]
        else:
            signing_keys = prod_cert

    for key in signing_keys["keys"]:
        if key["kid"] == unverified["header"]["kid"]:
            signing_key = key
            break

    algo = get_default_algorithms()["RS256"]  # type: ignore
    signing_key = algo.from_jwk(signing_key)
    verified = decode(
        token, key=signing_key, algorithms=["RS256"], options={"verify_aud": False}
    )

    return verified



from typing import Any, Dict, List

import requests

from fastel.payment.ecpay.models.invoice import IssueB2BModel, IssueB2CModel


class InvoiceSDK:
    PAYMENT_HOST = "https://payment-stg.revtel-api.com/v3"
    CLIENT_ID = ""
    CLIENT_SECRET = ""

    CARRIER_TYPE_TABLE = {
        "0": "",
        "1": "1",
        "2": "2",
        "3": "3",
    }

    def __init__(self, checkout: Dict[str, Any]):
        self.checkout = checkout

    def issue_invoice(self) -> Dict[str, Any]:
        if self.checkout.get("category", "B2C") == "B2C":
            url = f"{self.PAYMENT_HOST}/ecpay/B2C/invoice/issue?client_id={self.CLIENT_ID}&client_secret={self.CLIENT_SECRET}"
            invoice_data = self.generate_B2C_invoice_data()
        else:
            url = f"{self.PAYMENT_HOST}/ecpay/B2B/invoice/issue?client_id={self.CLIENT_ID}&client_secret={self.CLIENT_SECRET}"
            invoice_data = self.generate_B2B_invoice_data()
        resp = requests.post(url, json=invoice_data)
        print("issue invoice")
        resp_json = resp.json()
        assert isinstance(resp_json, dict)
        return resp_json

    @staticmethod
    def generate_B2C_invoice_item(
        items: List[Dict[str, Any]],
        extra_items: List[Dict[str, Any]],
        discount_items: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        result = []
        for item in items:
            item_result = {
                "ItemName": item.get("name", ""),
                "ItemCount": item["config"].get("qty", 1),
                "ItemWord": item["config"]["extra_data"]
                and item["config"]["extra_data"].get("word", "份")
                or "份",
                "ItemPrice": item.get("price", 0),
                "ItemAmount": item.get("amount", 0),
            }
            result.append(item_result)
        # 計算運費
        for extra_item in extra_items:
            item_result = {
                "ItemName": extra_item.get("name", ""),
                "ItemCount": 1,
                "ItemWord": "份",
                "ItemPrice": extra_item.get("amount", 0),
                "ItemAmount": extra_item.get("amount", 0),
            }
            result.append(item_result)
        # 計算折扣金額
        for discount_item in discount_items:
            item_result = {
                "ItemName": discount_item.get("name", ""),
                "ItemCount": 1,
                "ItemWord": "份",
                "ItemPrice": -discount_item.get("amount", 0),
                "ItemAmount": -discount_item.get("amount", 0),
            }
            result.append(item_result)
        return result

    def generate_B2C_invoice_data(self) -> Dict[str, Any]:
        items = self.checkout.get("items", [])
        extra_items = self.checkout.get("extra_items", [])
        discount_items = self.checkout.get("discount_items", [])
        carrier_type = self.CARRIER_TYPE_TABLE[
            self.checkout.get("invoice_carrier_type", "1")
        ]
        love_code = (
            self.checkout.get("invoice_donation", "0") == "1"
            and self.checkout.get("invoice_love_code", None)
            or None
        )
        carrier_num = (
            carrier_type in ["2", "3"]
            and self.checkout.get("invoice_carrier_num", "")
            or ""
        )
        invoice_data = {
            "RelateNumber": self.checkout.get("order_number", ""),
            "CustomerEmail": self.checkout.get("buyer_email", ""),
            "Print": "0",
            "Donation": self.checkout.get("invoice_donation", "0"),
            "LoveCode": love_code,
            "CarrierType": carrier_type,
            "CarrierNum": carrier_num,
            "TaxType": self.checkout.get("invoice_tax_type", "1"),
            "SalesAmount": self.checkout.get("total", 0),
            "Items": self.generate_B2C_invoice_item(items, extra_items, discount_items),
            "InvType": self.checkout.get("invoice_tax_type", "") == "4"
            and "08"
            or "07",
        }
        return IssueB2CModel.validate(invoice_data).dict(exclude_none=True)

    @staticmethod
    def generate_B2B_invoice_item(
        items: List[Dict[str, Any]],
        extra_items: List[Dict[str, Any]],
        discount_items: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        result = []
        seq = 1
        for item in items:
            item_result = {
                "ItemSeq": seq,
                "ItemName": item.get("name", ""),
                "ItemCount": item["config"].get("qty", 1),
                "ItemWord": item["config"]["extra_data"]
                and item["config"]["extra_data"].get("word", "份")
                or "份",
                "ItemPrice": item.get("unit_sales", 0),
                "ItemAmount": item.get("sales_amount", 0),
            }
            result.append(item_result)
            seq += 1
        # 計算運費
        for extra_item in extra_items:
            item_result = {
                "ItemSeq": seq,
                "ItemName": extra_item.get("name", ""),
                "ItemCount": 1,
                "ItemWord": "份",
                "ItemPrice": extra_item.get("sales_amount", 0),
                "ItemAmount": extra_item.get("sales_amount", 0),
            }
            result.append(item_result)
            seq += 1
        # 計算折扣金額
        for discount_item in discount_items:
            item_result = {
                "ItemSeq": seq,
                "ItemName": discount_item.get("name", ""),
                "ItemCount": 1,
                "ItemWord": "份",
                "ItemPrice": -discount_item.get("sales_amount", 0),
                "ItemAmount": -discount_item.get("sales_amount", 0),
            }
            result.append(item_result)
            seq += 1
        return result

    def generate_B2B_invoice_data(self) -> Dict[str, Any]:
        items = self.checkout.get("items", [])
        extra_items = self.checkout.get("extra_items", [])
        discount_items = self.checkout.get("discount_items", [])
        invoice_data = {
            "RelateNumber": self.checkout.get("order_number", ""),
            "CustomerIdentifier": self.checkout.get("invoice_uni_no", ""),
            "CustomerEmail": self.checkout.get("buyer_email", ""),
            "TaxType": self.checkout.get("invoice_tax_type", "1"),
            "SalesAmount": self.checkout.get("sales", 0),
            "TaxAmount": self.checkout.get("tax", 0),
            "TotalAmount": self.checkout.get("total", 0),
            "Items": self.generate_B2B_invoice_item(items, extra_items, discount_items),
            "InvType": self.checkout.get("invoice_tax_type", "") == "4"
            and "08"
            or "07",
        }
        return IssueB2BModel.validate(invoice_data).dict(exclude_none=True)



from typing import Any, Tuple


def prod_cert(*args: Any, **kwargs: Any) -> Tuple[bool, Any]:
    return (False, None)


def stg_cert(*args: Any, **kwargs: Any) -> Tuple[bool, Any]:
    return (False, None)
